#!/bin/bash

set -e

usage() {
    cat <<END_USAGE
Usage: $PROG IMAGE DEVICE
\`$PROG' populates a USB stick with the Ubuntu installer.

DEVICE should name a partition of a USB disk (e.g. \`/dev/sdb1') or a
target image file. If it's an image file, \`$PROG' will prepare a DVD or
USB disk image depending on the file name. If it ends in \`.iso', \`$PROG'
will prepare a bootable ISO image which can be written to a DVD. Otherwise,
it will prepare a USB disk image which can be written to a USB stick with
\`dd' or a similar program.

 Options:
      --help                 display this message
      --aes-dir=DIRECTORY    use DIRECTORY as the source of AES
                             customizations, defaults to the location of this
                             program
      --aes-repo=DIRECTORY   use DIRECTORY as the AES package repository,
                             defaults to the location of this program
      --country=CC           preseed \`mirror/country' with given country code,
                             defaults to \`US'
      --force                unconditionally overwrite the target image file
      --keyboard=KEYBOARD    preseed \`kbd-chooser/method' and
                             \`console-setup/layoutcode' with KEYBOARD,
                             defaults to \`us'
      --locale=LOCALE        preseed \`debian-installer/locale' with LOCALE,
                             defaults to \`en_US'
      --no-format            don't format the target device
      --no-mbr               don't install a master boot record
      --no-syslinux          don't install the syslinux boot loader
      --update               update the USB stick; implies --no-format,
                             --no-syslinux, and --no-mbr
  -q                         be quiet
  -v                         be more verbose
END_USAGE
}

PROG=${0##*/}

LOG_INFO=6
OPT_AES_DIR=${0%/*}
OPT_AES_REPO=$OPT_AES_DIR
OPT_COUNTRY=US
OPT_VERBOSE=$LOG_INFO
OPT_FORCE=false
OPT_FORMAT=true
OPT_SYSLINUX=true
OPT_MBR=true
OPT_KEYBOARD=us
OPT_LOCALE=en_US
OPT_UPDATE=false

IMAGE=
DEVICE=
SOURCE=
SUITE=
ARCH=
TARGET=
TARGET_DISK_IMAGE=
TARGET_ISO_IMAGE=
TEMPFILE=

SYSLINUX_TYPE='syslinux'
SYSLINUX_TARGET_DIR=
SYSLINUX_LIBDIR='/usr/lib/syslinux'
SYSLINUX_GFXBOOT_C32="$SYSLINUX_LIBDIR/gfxboot.c32"
SYSLINUX_VESAMENU_C32="$SYSLINUX_LIBDIR/vesamenu.c32"
SYSLINUX_MBR="$SYSLINUX_LIBDIR/mbr.bin"
SYSLINUX_ISOLINUX_BIN="$SYSLINUX_LIBDIR/isolinux.bin"

# Required commands that might not be installed.
REQUIRED_COMMANDS=()
i=0
REQUIRED_COMMANDS[i++]='fakeroot'
REQUIRED_COMMANDS[i++]='genisoimage'
REQUIRED_COMMANDS[i++]='lsb_release'
REQUIRED_COMMANDS[i++]='mkdosfs'
REQUIRED_COMMANDS[i++]='rsync'
REQUIRED_COMMANDS[i++]='sfdisk'
REQUIRED_COMMANDS[i++]='syslinux'

# Required files that might not be installed.
REQUIRED_FILES=()
i=0
REQUIRED_FILES[i++]="$SYSLINUX_GFXBOOT_C32"
REQUIRED_FILES[i++]="$SYSLINUX_VESAMENU_C32"
REQUIRED_FILES[i++]="$SYSLINUX_MBR"
REQUIRED_FILES[i++]="$SYSLINUX_ISOLINUX_BIN"

main() {
    parse_command_line ${1+"$@"}
    check_prerequisites
    mount_source
    ! $OPT_FORMAT || [ -z "$TARGET_DISK_IMAGE" ] || make_target_disk_image
    [ -z "$TARGET_DISK_IMAGE" ] || setup_target_device
    ! $OPT_FORMAT || format_target
    mount_target
    copy_files
    customize
    ! $OPT_SYSLINUX || install_syslinux
    ! $OPT_MBR || install_mbr
    [ -z "$TARGET_ISO_IMAGE" ] || make_target_iso_image
}

parse_command_line() {
    local args
    args=$(getopt -n "$PROG"		\
                  -o 'qv'		\
                  -l 'aes-dir:'		\
                  -l 'aes-repo:'	\
                  -l 'country:'		\
                  -l 'force'		\
                  -l 'help'		\
                  -l 'keyboard:'	\
                  -l 'locale:'		\
                  -l 'no-format'	\
                  -l 'no-mbr'		\
                  -l 'no-syslinux'	\
                  -l 'update'		\
		  -- "$@")		\
	|| { usage >&2; exit 1; }
    eval set -- "$args"
    while :; do
	case "$1" in
	  --aes-dir)
		OPT_AES_DIR=$2; shift;;
	  --aes-repo)
		OPT_AES_REPO=$2; shift;;
	  --country)
		OPT_COUNTRY=$2; shift;;
	  --force)
		OPT_FORCE=true;;
	  --keyboard)
		OPT_KEYBOARD=$2; shift;;
	  --locale)
		OPT_LOCALE=$2; shift;;
	  --no-format)
		OPT_FORMAT=false;;
	  --no-mbr)
		OPT_MBR=false;;
	  --no-syslinux)
		OPT_SYSLINUX=false;;
	  --update)
		OPT_UPDATE=true;;
	  -q)	OPT_VERBOSE=0;;
	  -v)   (( ++OPT_VERBOSE ));;
	  -\?|--help)
        	usage; exit 0;;
	  --)	shift; break;;
	  *)	echo 'Internal error!'; exit 1;;
	esac
	shift
    done

    IMAGE=$1; DEVICE=$2; : ${IMAGE:?} ${DEVICE:?}

    if [ -b "$DEVICE" ]; then
	:
    elif [ -f "$DEVICE" -o ! -e "$DEVICE" ]; then
	if [[ "$DEVICE" == *.iso ]]; then
	    TARGET_ISO_IMAGE=$DEVICE
	else
	    TARGET_DISK_IMAGE=$DEVICE
	fi
	DEVICE=
    else
	echo "\`$DEVICE' is not a block device or regular file" >&2; exit 1
    fi

    if [ -n "$TARGET_ISO_IMAGE" ]; then
	OPT_FORMAT=false
	OPT_MBR=false
	OPT_SYSLINUX=true
	OPT_UPDATE=false
	SYSLINUX_TYPE=isolinux
    elif $OPT_UPDATE; then
	OPT_FORMAT=false
	OPT_SYSLINUX=false
	OPT_MBR=false
    fi
}

check_prerequisites() {
    local rv=0 x
    for x in ${REQUIRED_COMMANDS[*]}; do
	if ! type -p $x >/dev/null; then
	    echo "$x is not installed" >&2
	    rv=1
	fi
    done
    for x in "${REQUIRED_FILES[@]}"; do
	if ! [ -r "$x" ]; then
	    echo "File \`$x' doesn't exist or is not readable" >&2
	    rv=1
	fi
    done
    return $rv
}

cleanup() {
    set +e
    [ -z "$TEMPFILE" ] || rm -f "$TEMPFILE"
    if [ -n "$TARGET_ISO_IMAGE" -a -d "$TARGET" ]; then
	log_info "Removing staging tree \`$TARGET'"
	rm -rf "$TARGET"
    else
	umount_and_remove TARGET
    fi
    [[ "$DEVICE" == /dev/loop[0-9]* ]] && losetup -d "$DEVICE"
    umount_and_remove SOURCE
}

make_target_disk_image() {
    local file=${TARGET_DISK_IMAGE:?}
    log_info "Making target image file \`$file'"
    if [ -e "$file" ]; then
	if ! $OPT_FORCE; then
	    read -p "overwrite \`$file'? [y/N] "
	    [[ "$REPLY" == [yY]* ]] || exit 0
	fi
	rm "$file"
    fi
    make_file "$file" $(( 2000 * 1000 * 1000 ))
    partition_target_disk_image
}

make_file() {	# image size
    local image=$1 size=$2
    local block_count=$(( ( size + 511 ) / 512 ))
    TEMPFILE=`mktemp`
    dd if=/dev/zero of="$image" seek=$(( block_count - 1 )) count=1 \
      >>"$TEMPFILE" 2>&1 || {
	status=$?
	cat "$TEMPFILE" >&2
	exit $status
    }
    rm "$TEMPFILE"; TEMPFILE=
}

partition_target_disk_image() {
    log_info "Partitioning target image file"
    TEMPFILE=`mktemp`
    # <start> <size> <id> <bootable>
    sfdisk -D -H 255 -S 63 -f -q "$TARGET_DISK_IMAGE" <<-'END_PARTITION_TABLE' \
      >>"$TEMPFILE" 2>&1
	,,c,*
	;
	;
	;
	END_PARTITION_TABLE
    local status=$?
    if [ $status -ne 0 ]; then
	cat "$TEMPFILE" >&2
	exit $status
    fi
    rm "$TEMPFILE"; TEMPFILE=
}

make_target_iso_image() {
    log_info "Generating ISO image \`$TARGET_ISO_IMAGE'"
    genisoimage -quiet -o "$TARGET_ISO_IMAGE" -r -J \
	-b isolinux/isolinux.bin \
	-c isolinux/boot.cat \
	-no-emul-boot \
	-boot-load-size 4 \
	-boot-info-table \
	"$TARGET"
}

setup_target_device() {
    DEVICE=`losetup -fs -o $(( 63 * 512 )) "$TARGET_DISK_IMAGE"`
}

mount_source() {
    SOURCE=$(mktemp -d ${TMPDIR:-/tmp}/source.XXXXXXXXXX)
    log_info "Mounting source on \`$SOURCE'"
    mount -r -o loop "$IMAGE" "$SOURCE"
    SUITE=`sed -ne 's/^Ubuntu[^"]\+"\([^[:space:]"]\+\).*$/\1/p' \
	"$SOURCE/.disk/info"`
    SUITE=`lc "$SUITE"`
    ARCH=`sed -ne 's/^.*\b\(i386\|amd64\)\b.*$/\1/p' "$SOURCE/.disk/info"`
}

mount_target() {
    TARGET=$(mktemp -d ${TMPDIR:-/tmp}/target.XXXXXXXXXX)
    SYSLINUX_TARGET_DIR="$TARGET/$SYSLINUX_TYPE"
    if [ -n "$TARGET_ISO_IMAGE" ]; then
	log_info "Creating staging tree in \`$TARGET'"
    else
	log_info "Mounting target on \`$TARGET'"
	mount "$DEVICE" "$TARGET"
    fi
}

umount_and_remove() {
    local var=$1 dir
    eval dir=\$$var
    if [ -n "$dir" ]; then
	if is_mounted "$dir"; then
	    log_info "Unmounting `lc $var`"
	    umount "$dir"
	fi
	! [ -d "$dir" ] || rmdir "$dir"
    fi
}

copy_files() {
    local c=''
    ! $OPT_UPDATE || c='c'
    log_info "Copying installer"
    rsync -${c}aO --no-l --no-p --no-o --no-g \
	--exclude '/isolinux/' \
	"$SOURCE/" "$TARGET"
    rsync -${c}aO --no-l --no-p --no-o --no-g \
	--exclude '/boot.cat' \
	--exclude '/isolinux.bin' \
	--exclude '/isolinux.cfg' \
	--exclude '*.c32' \
	"$SOURCE/isolinux/" "$SYSLINUX_TARGET_DIR"

    # Accomodate varying support and syntax of syslinux graphical menus
    # (a.k.a `gfxboot').
    local off ui
    case "`lsb_release -s -c`" in
      lenny)	off='# '; ui='\1';;	# Disable `gfxboot'.
      lucid)	off=''  ; ui='';;	# Use syntax `gfxboot ...'.
      *)	off=''  ; ui='\1';;	# Use syntax `ui gfxboot ...'.
    esac
    sed -e "s/^\\(ui[[:space:]]\\+\\)\\?\\(gfxboot.*\\)\$/$off$ui\\2/" \
	"$SOURCE/isolinux/isolinux.cfg" \
	>"$SYSLINUX_TARGET_DIR/$SYSLINUX_TYPE.cfg"
}

copy_repo() {
    local c=''
    ! $OPT_UPDATE || c='c'
    log_info "Copying AES package repository"
    rsync -${c}aRO --no-l --no-p --no-o --no-g \
	--include "/dists/$SUITE/Contents-$ARCH*" \
	--exclude "/dists/$SUITE/Contents-*" \
	--include "binary-$ARCH/" \
	--exclude "binary-*/" \
	--exclude '*.gpg' \
 	"$OPT_AES_REPO/./dists/$SUITE" "$TARGET/aes"
}

customize() {
    copy_repo

    local c=''
    ! $OPT_UPDATE || c='c'
    log_info "Copying AES customizations"
    rsync -${c}aO --no-l --no-p --no-o --no-g \
	--exclude '/dists/' \
	--exclude '/hooks/' \
	--exclude '/hooks.sh' \
	--exclude '/modules/' \
	--exclude '*~' \
	--exclude '*.swp' \
	"$OPT_AES_DIR/" "$TARGET/aes"
    fakeroot tar cfz "$TARGET/aes/hooks.tgz" -C "$OPT_AES_DIR" hooks.sh hooks
    fakeroot tar cfz "$TARGET/aes/modules.tgz" -C "$OPT_AES_DIR/modules/$ARCH" .

    # Customize preseed configuration file.

    local keyboard_param
    case "$SUITE" in
      lucid|maverick)   keyboard_param='console-setup/layoutcode';;
      *)                keyboard_param='keyboard-configuration/layoutcode';;
    esac

    awk '
	BEGIN {
	    param["mirror/country"]		= "'$OPT_COUNTRY'"
	    param["console-setup/layoutcode"]	= "'$OPT_KEYBOARD'"
	    param["debian-installer/locale"]	= "'$OPT_LOCALE'"
        }
        $1 == "d-i" && $2 in param { $4 = param[$2] }
        $1 == "d-i" && $2 == "console-setup/layoutcode" {
	    $2 = "'$keyboard_param'"
	}
        { print }
    ' <"$TARGET"/aes/preseed.cfg >"$TARGET"/aes/preseed.$$
    mv "$TARGET"/aes/preseed.{$$,cfg}

    # Customize syslinux configuration.

    cat >"$SYSLINUX_TARGET_DIR/aes.cfg" <<-END_SYSLINUX_AES_CFG
	label auto
	  menu label Install Ubuntu ^Automatically
	  kernel /install/vmlinuz
	  append noprompt auto file=/cdrom/aes/./preseed.cfg debian-installer/locale=$OPT_LOCALE kbd-chooser/method=$OPT_KEYBOARD $keyboard_param=$OPT_KEYBOARD cdrom-detect/try-usb=true initrd=/install/initrd.gz quiet --
	END_SYSLINUX_AES_CFG

    local text=`echo "$SYSLINUX_TARGET_DIR"/t*xt.cfg`
    text="${text%.cfg}"
    awk '
	BEGIN			{ stanza = 0 }
	/^label/ && stanza == 1	{ print "include aes.cfg" }
	/^label/                { ++stanza }
                                { print }
    ' <"$text.cfg" >"$text.$$"
    mv "$text".{$$,cfg}
}

format_target() {
    log_info "Formatting \`$DEVICE'"
    mkdosfs -F 32 "$DEVICE"
}

install_syslinux() {
    cp -p "$SYSLINUX_GFXBOOT_C32"  "$SYSLINUX_TARGET_DIR"
    cp -p "$SYSLINUX_VESAMENU_C32" "$SYSLINUX_TARGET_DIR"
    cp -p "$SYSLINUX_ISOLINUX_BIN" "$SYSLINUX_TARGET_DIR"
    [ -z "$TARGET_ISO_IMAGE" ] || return 0
    umount_and_remove TARGET
    log_info "Installing syslinux boot loader to \`$DEVICE'"
    syslinux "$DEVICE"
}

install_mbr() {
    umount_and_remove TARGET
    local disk="${TARGET_DISK_IMAGE:-${DEVICE%%[0-9]?([0-9])}}"
    log_info "Installing master boot record to \`$disk'"
    TEMPFILE=`mktemp`
    dd if=/usr/lib/syslinux/mbr.bin of="$disk" conv=notrunc 2>>"$TEMPFILE" || {
	status=$?
	cat "$TEMPFILE" >&2
	exit $status
    }
}

is_mounted() {
    tac /proc/mounts | awk '
        BEGIN           { found = 0 }
        $2 == "'$1'"    { found = 1; exit }
        END             { exit !found }
    '
    return $?
}

lc() {
    echo "$1" | tr '[:upper:]' '[:lower:]'
}

log_info() {
    if [ $# -gt 0 -a $OPT_VERBOSE -ge $LOG_INFO ]; then
	echo "$@"
    fi
}

trap cleanup 0 1 2 15

shopt -s extglob

main ${1+"$@"}

exit 0

